# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2019 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Handlers for customizing oauthclient endpoints."""

from functools import partial, wraps

from flask import (
    current_app,
    flash,
    redirect,
    render_template,
    request,
    session,
    url_for,
)
from flask_login import current_user
from invenio_db import db
from invenio_i18n import gettext as _

from ..errors import (
    AlreadyLinkedError,
    OAuthClientAlreadyAuthorized,
    OAuthClientError,
    OAuthClientMustRedirectLogin,
    OAuthClientMustRedirectSignup,
    OAuthClientTokenNotFound,
    OAuthClientTokenNotSet,
    OAuthClientUnAuthorized,
    OAuthClientUserNotRegistered,
    OAuthError,
    OAuthRejectedRequestError,
)
from ..utils import create_registrationform, fill_form
from .authorized import authorized_handler, extra_signup_handler
from .base import base_disconnect_handler
from .decorators import can_extra_signup
from .token import response_token_setter, token_session_key


def _oauth_error_handler(remote, f, *args, **kwargs):
    """Function to handle exceptions."""
    try:
        return f(remote, *args, **kwargs)
    except OAuthClientError as e:
        current_app.logger.warning(e.message, exc_info=True)
        return oauth2_handle_error(e.remote, e.response, e.code, e.uri, e.description)
    except OAuthClientUnAuthorized:
        return current_app.login_manager.unauthorized()
    except AlreadyLinkedError:
        flash(
            _("External service is already linked to another account."),
            category="danger",
        )
        return redirect(url_for("invenio_oauthclient_settings.index"))
    except OAuthRejectedRequestError:
        flash(_("You rejected the authentication request."), category="info")
        return redirect("/")
    except OAuthClientAlreadyAuthorized:
        return redirect("/")
    except OAuthClientTokenNotFound:
        return redirect("/")
    except OAuthClientUserNotRegistered:
        raise OAuthError("Could not create user.", remote)
    except OAuthClientTokenNotSet:
        raise OAuthError("Could not create token for user.", remote)
    except OAuthClientMustRedirectSignup as e:
        return redirect(
            url_for(
                ".signup",
                remote_app=remote.name,
            )
        )
    except OAuthClientMustRedirectLogin as e:
        return redirect(
            url_for(
                ".login",
                remote_app=remote.name,
            )
        )


#
# Error handling decorators
#
def oauth_resp_remote_error_handler(f):
    """Decorator to handle exceptions."""

    @wraps(f)
    def inner(resp, remote, *args, **kwargs):
        # OAuthErrors should not happen, so they are not caught here. Hence
        # they will result in a 500 Internal Server Error which is what we
        # are interested in.
        _f = partial(f, resp)
        return _oauth_error_handler(remote, _f, *args, **kwargs)

    return inner


def oauth_remote_error_handler(f):
    """Decorator to handle exceptions."""

    @wraps(f)
    def inner(remote, *args, **kwargs):
        # OAuthErrors should not happen, so they are not caught here. Hence
        # they will result in a 500 Internal Server Error which is what we
        # are interested in.
        return _oauth_error_handler(remote, f, *args, **kwargs)

    return inner


#
# Handlers
#
@oauth_resp_remote_error_handler
def authorized_default_handler(resp, remote, *args, **kwargs):
    """Store access token in session.

    Default authorized handler.

    :param remote: The remote application.
    :param resp: The response.
    :returns: Redirect response.
    """
    response_token_setter(remote, resp)
    db.session.commit()
    return redirect(url_for("invenio_oauthclient_settings.index"))


@oauth_resp_remote_error_handler
def authorized_signup_handler(resp, remote, *args, **kwargs):
    """Handle sign-in/up functionality.

    :param remote: The remote application.
    :param resp: The response.
    :returns: Redirect response.
    """
    next_url = authorized_handler(resp, remote, *args, **kwargs)
    # Redirect to next
    if next_url:
        return redirect(next_url)
    return redirect(url_for("invenio_oauthclient_settings.index"))


@oauth_remote_error_handler
def disconnect_handler(remote, *args, **kwargs):
    """Handle unlinking of remote account.

    This default handler will just delete the remote account link. You may
    wish to extend this module to perform clean-up in the remote service
    before removing the link (e.g. removing install webhooks).

    :param remote: The remote application.
    :returns: Redirect response.
    """
    base_disconnect_handler(remote, *args, **kwargs)
    return redirect(url_for("invenio_oauthclient_settings.index"))


@oauth_remote_error_handler
@can_extra_signup
def signup_handler(remote, *args, **kwargs):
    """Handle extra signup information.

    The info in the request.form is used to fill in the user registration form.
    When the form does not validate (e.g. missing user e-mail), it is displayed
    to the user, who needs to fill in the missing info and submit it.
    This method is also called when the user has filled in the form and click on
    btn submit.

    :param remote: The remote application.
    :returns: Redirect response or the template rendered.
    """
    form = create_registrationform(request.form, oauth_remote_app=remote)
    if not form.is_submitted():
        session_prefix = token_session_key(remote.name)
        account_info = session.get(session_prefix + "_account_info")
        # Pre-fill form
        fill_form(form, account_info["user"])
        return render_template(
            current_app.config["OAUTHCLIENT_SIGNUP_TEMPLATE"],
            form=form,
            remote=remote,
            app_title=current_app.config["OAUTHCLIENT_REMOTE_APPS"][remote.name].get(
                "title", ""
            ),
            app_description=current_app.config["OAUTHCLIENT_REMOTE_APPS"][
                remote.name
            ].get("description", ""),
            app_icon=current_app.config["OAUTHCLIENT_REMOTE_APPS"][remote.name].get(
                "icon", None
            ),
        )
    elif not form.errors:
        try:
            next_url = extra_signup_handler(remote, form, *args, **kwargs)
        except OAuthClientUnAuthorized:
            # Redirect the user after registration (which doesn't include the
            # activation), waiting for user to confirm his email.
            return redirect(url_for("security.login"))

        if next_url:
            return redirect(next_url)
        else:
            return redirect("/")


def oauth2_handle_error(remote, resp, error_code, error_uri, error_description):
    """Handle errors during exchange of one-time code for an access tokens."""
    flash(_("Authorization with remote service failed."))
    return redirect("/")



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Blueprints for oauthclient."""

from .client import blueprint as client_blueprint
from .settings import blueprint as settings_blueprint

blueprints = [
    client_blueprint,
    settings_blueprint,
]



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2023 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Decorators."""

from functools import wraps

from flask import session
from flask_login import current_user

from ..errors import (
    OAuthClientAlreadyAuthorized,
    OAuthClientMustRedirectLogin,
    OAuthClientTokenNotFound,
)
from .token import token_getter, token_session_key


def can_extra_signup(f):
    """Ensure that the extra signup handler can be called.

    The handler can be called when the user is not yet authenticated
    and when the OAuth token is valid.
    """

    @wraps(f)
    def wrapper(*args, **kwargs):
        # make sure the user is not already authenticated
        if current_user.is_authenticated:
            raise OAuthClientAlreadyAuthorized()

        remote = args[0]
        # Make sure that there is an OAuth token in the session, to ensure that
        # we are in the OAuth flow
        oauth_token = token_getter(remote)
        if not oauth_token:
            raise OAuthClientTokenNotFound()

        # Make sure that `authorized` step executed first, and that the extra signup
        # step is actually needed because some user info is missing
        session_prefix = token_session_key(remote.name)
        if not session.get(session_prefix + "_autoregister", False):
            raise OAuthClientMustRedirectLogin()

        return f(*args, **kwargs)

    return wrapper



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2023 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Authorize handlers."""

from flask import current_app, g, session
from flask_login import current_user
from invenio_db import db

from ..errors import (
    OAuthClientAlreadyAuthorized,
    OAuthClientMustRedirectSignup,
    OAuthClientTokenNotFound,
    OAuthClientTokenNotSet,
    OAuthClientUnAuthorized,
    OAuthClientUserNotRegistered,
)
from ..oauth import oauth_authenticate, oauth_get_user, oauth_register
from ..proxies import current_oauthclient
from ..signals import (
    account_info_received,
    account_setup_committed,
    account_setup_received,
)
from ..utils import create_csrf_disabled_registrationform, fill_form
from .token import (
    get_session_next_url,
    response_token_setter,
    token_getter,
    token_session_key,
    token_setter,
)
from .utils import create_or_update_roles


def authorized_handler(resp, remote, *args, **kwargs):
    """Handle user login after OAuth authorize step.

    :param resp: The response of the `authorized` endpoint.
    :param remote: The remote application.
    :returns: The URL to go next after login
    """
    # Validate the response and set token in the user session. This must happen
    # first to make sure that the response payload is valid.
    # Returned token is None when anonymous user
    token = response_token_setter(remote, resp)

    # Set the remote in the user session to know how the user logged in.
    # Useful on log out, so that we can logout on remote too, when needed.
    session["OAUTHCLIENT_SESSION_REMOTE_NAME"] = remote.name
    # Remove any previously stored auto register session key
    session.pop(token_session_key(remote.name) + "_autoregister", None)

    handlers = current_oauthclient.signup_handlers[remote.name]

    # call user info endpoint
    account_info = handlers["info"](resp)
    assert "external_id" in account_info
    account_info_received.send(remote, response=resp, account_info=account_info)

    # call groups endpoint, when defined
    group_handler = handlers.get("groups")
    if group_handler:
        account_groups = group_handler(resp)
        if account_groups:
            roles_ids = create_or_update_roles(account_groups)
            # Set the unmanaged roles in the user session, used in other modules.
            # Unmanaged user roles are not stored in the DB for privacy reasons:
            # sys admins should not know the external groups of a user.
            session["unmanaged_roles_ids"] = roles_ids

    # In the normal OAuth flow, the user is not yet authenticated. However, it the user
    # is already logged in, and goes to 'Linked accounts', clicks 'Connect' on another
    # remote app, `authorized` will be called with the new remote.
    is_normal_oauth_flow = not current_user.is_authenticated
    if is_normal_oauth_flow:
        # get the user from the DB using the current remote
        user = oauth_get_user(
            remote.consumer_key,
            account_info=account_info,
            access_token=token_getter(remote)[0],
        )
        if user is None:
            # User not found, this is the first login. Register the user.
            # The registration raises an exception when the account info is not enough
            # to register the user.
            form = create_csrf_disabled_registrationform(remote)
            form = fill_form(form, account_info["user"])

            try:
                user = _register_user(resp, remote, account_info, form)
            except OAuthClientUserNotRegistered:
                # save in the session info to display the extra signup form to the user
                session[token_session_key(remote.name) + "_autoregister"] = True
                session[token_session_key(remote.name) + "_account_info"] = account_info
                session[token_session_key(remote.name) + "_response"] = resp
                db.session.commit()
                # this will trigger a redirect to /signup (therefor
                # signup_handler/extra_signup_handler funcs) and will require the user
                # to fill in the registration form with the missing information
                raise OAuthClientMustRedirectSignup()

        if not oauth_authenticate(
            remote.consumer_key, user, require_existing_link=False
        ):
            raise OAuthClientUnAuthorized()
        # Store token in the database instead of only the session
        token = response_token_setter(remote, resp)

    return _complete_authorize(resp, remote, handlers, token)


def _register_user(resp, remote, account_info, form):
    """Try to register the user with info got from the remote app.

    :param resp: The response of the `authorized` endpoint.
    :param remote: The remote application.
    """
    remote_app = current_app.config["OAUTHCLIENT_REMOTE_APPS"][remote.name]
    precedence_mask = remote_app.get("precedence_mask")
    signup_options = remote_app.get("signup_options")

    user = oauth_register(
        form,
        account_info["user"],
        precedence_mask=precedence_mask,
        signup_options=signup_options,
    )

    if user is None:
        # Registration failed: the account info is not enough to register the user.
        # Save info in the session, necessary for the user registration flow with form.
        raise OAuthClientUserNotRegistered()

    return user


def _complete_authorize(resp, remote, handlers, token):
    """Complete authorized flow.

    This happens after:
     - A normal authorized flow.
     - The extra signup registration, where the user needs to fill in the missing
       information during the first login.
    """
    is_first_login_with_this_remote = not token.remote_account.extra_data
    if is_first_login_with_this_remote:
        # call the `setup` handler to get complete the first login with this remote
        account_setup = handlers["setup"](token, resp)
        account_setup_received.send(
            remote, token=token, response=resp, account_setup=account_setup
        )
        db.session.commit()
        account_setup_committed.send(remote, token=token)
    else:
        db.session.commit()

    # Return the URL where to go next
    next_url = get_session_next_url(remote.name)
    if next_url:
        return next_url


def extra_signup_handler(remote, form, *args, **kwargs):
    """Handle extra signup information.

    Validate the extra missing information that the user inserted in the signup step,
    after authorized. If the info inserted are enough

    :param remote: The remote application.
    :returns: Redirect response or the template rendered.
    """
    if current_user.is_authenticated:
        raise OAuthClientAlreadyAuthorized()

    # Retrieve token from session
    oauth_token = token_getter(remote)
    if not oauth_token:
        raise OAuthClientTokenNotFound()

    session_prefix = token_session_key(remote.name)

    handlers = current_oauthclient.signup_handlers[remote.name]
    if form.validate_on_submit():
        # remove the autoregister flag that marks the authorized flow
        session.pop(session_prefix + "_autoregister", None)
        account_info = session.pop(session_prefix + "_account_info")
        response = session.pop(session_prefix + "_response")

        user = _register_user(response, remote, account_info, form)

        # Link account and set session data
        token = token_setter(remote, oauth_token[0], secret=oauth_token[1], user=user)
        if token is None:
            raise OAuthClientTokenNotSet()

        # Authenticate user, without requiring the existence of the RemoteAccount,
        # which is created later in the setup handler.
        if not oauth_authenticate(
            remote.consumer_key, user, require_existing_link=False
        ):
            raise OAuthClientUnAuthorized()

        return _complete_authorize(response, remote, handlers, token)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2023 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""OAuth methods to help find, authenticate or register a remote user."""

from flask import after_this_request, current_app
from flask_security import login_user, logout_user
from flask_security.confirmable import confirm_user, requires_confirmation
from invenio_accounts.models import User
from invenio_accounts.utils import register_user
from werkzeug.local import LocalProxy

from .models import RemoteAccount, RemoteToken, UserIdentity
from .utils import (
    fill_form,
    filter_user_info,
    patch_dictionary,
    remove_csrf_tokens,
    remove_none_values,
)

_security = LocalProxy(lambda: current_app.extensions["security"])

_datastore = LocalProxy(lambda: _security.datastore)


def _commit(response=None):
    _datastore.commit()
    return response


def _get_external_id(account_info):
    """Get external id from account info."""
    if all(k in account_info for k in ("external_id", "external_method")):
        return dict(
            id=account_info["external_id"], method=account_info["external_method"]
        )
    return None


def oauth_get_user(client_id, account_info=None, access_token=None):
    """Retrieve user object for the given request.

    Uses either the access token or extracted account information to retrieve
    the user object.

    :param client_id: The client id.
    :param account_info: The dictionary with the account info.
        (Default: ``None``)
    :param access_token: The access token. (Default: ``None``)
    :returns: A :class:`invenio_accounts.models.User` instance or ``None``.
    """
    if access_token:
        token = RemoteToken.get_by_token(client_id, access_token)
        if token:
            return token.remote_account.user

    if account_info:
        external_id = _get_external_id(account_info)
        if external_id:
            user = UserIdentity.get_user(external_id["method"], external_id["id"])
            if user:
                return user
        email = account_info.get("user", {}).get("email")
        if email:
            return User.query.filter_by(email=email).one_or_none()
    return None


def oauth_authenticate(client_id, user, require_existing_link=False):
    """Authenticate an oauth authorized callback.

    :param client_id: The client id.
    :param user: A user instance.
    :param require_existing_link: If ``True``, check if remote account exists.
        (Default: ``False``)
    :returns: ``True`` if the user is successfully authenticated.
    """
    # Authenticate via the access token (access token used to get user_id)
    if not requires_confirmation(user):
        after_this_request(_commit)
        if login_user(user):
            if require_existing_link:
                account = RemoteAccount.get(user.id, client_id)
                if account is None:
                    logout_user()
                    return False
            return True
    return False


def oauth_register(form, user_info=None, precedence_mask=None, signup_options={}):
    """Register user if possible.

    :param form: A form instance.
    :param user_info: The user info dictionary.
    :param precedence_mask: The precedence mask to use.
    :param signup_options: Extra signup options dict.
    :returns: A :class:`invenio_accounts.models.User` instance.
    """
    if form.validate():
        form_data = form.data

        # let relevant information from the OAuth service's user info
        # have precedence over the values specified by the user
        if user_info:
            default_mask = {"email": True}
            # act on form data so the `profile` is updated correctly
            filter_user_info(user_info, precedence_mask or default_mask)
            patch_dictionary(form_data, user_info)

        # re-populate the form after applying the precedence mask and
        # convert the form data to user model's data
        data = fill_form(form, form_data).to_dict()
        # see https://github.com/inveniosoftware/invenio-oauthclient/issues/275
        remove_none_values(data)
        # remove the CSRF tokens to avoid unexpected keyword arguments
        remove_csrf_tokens(data)

        send_register_msg = signup_options.get("send_register_msg", True)
        user = register_user(send_register_msg=send_register_msg, **data)
        auto_confirm = signup_options.get("auto_confirm", False)
        if auto_confirm:
            confirm_user(user)
        _datastore.commit()
        return user


def oauth_link_external_id(user, external_id=None):
    """Link a user to an external id.

    :param user: A :class:`invenio_accounts.models.User` instance.
    :param external_id: The external id associated with the user.
        (Default: ``None``)
    :raises invenio_oauthclient.errors.AlreadyLinkedError: Raised if already
        exists a link.
    """
    # Backward compatibility. Use UserIdentity directly instead of this method.
    UserIdentity.create(user, external_id["method"], external_id["id"])


def oauth_unlink_external_id(external_id):
    """Unlink a user from an external id.

    :param external_id: The external id associated with the user.
    """
    UserIdentity.delete_by_external_id(external_id["method"], external_id["id"])
